
/*-------------------------------------------------------------------*/
/* MSTKruskal.c    ZHONG Caiming                                     */
/*                 07-08-2009                                        */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "MSTKruskal"
#define VersionNumber  "Version 0.1"
#define LastUpdated    "26.08.2009"

/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "MSTKruskal.h"
#include "memctrl.h"


/*-------------------------------------------------------------------*/

void EuclideanDistFF(float * vector1, 
                     float * vector2, 
                     int     vectorSize,
                     float * f)
{
  int i;
  *f = 0;

  for (i = 0; i < vectorSize; i ++){
    *f += (vector1[i] - vector2[i]) * 
          (vector1[i] - vector2[i]);
  }
  *f = sqrt(*f);
}

/*-------------------------------------------------------------------*/

void EuclideanDistIF(int    * vector1, 
                     float  * vector2, 
                     int      vectorSize,
                     float  * f)
{
  int i;
  *f = 0;

  for (i = 0; i < vectorSize; i ++){
    *f += ((float)vector1[i] - vector2[i]) * 
          ((float)vector1[i] - vector2[i]);
  }
  *f = sqrt(*f);
}

/*-------------------------------------------------------------------*/

void EuclideanDistII(int    * vector1, 
                     int    * vector2, 
                     int      vectorSize,
                     float  * f)
{
  int i;
  *f = 0;

  for (i = 0; i < vectorSize; i ++){
    *f += ((float)vector1[i] - (float)vector2[i]) * 
          ((float)vector1[i] - (float)vector2[i]);
  }
  *f = sqrt(*f);
}
                      
/*-------------------------------------------------------------------*/
/* To produce pairs (edges) of the complete graph
   Pair<weight, index1, index2>                                      */ 

PairInfo * OrganizePairsFromTrainingSet(TRAININGSET *TS)
{
  int         bookSize = BookSize(TS);
  int         i, j, idx;

  Pair      * Pairs = (Pair*)allocate(sizeof(Pair) * bookSize * (bookSize - 1) / 2);
  PairInfo  * PInfo = (PairInfo*)allocate(sizeof(PairInfo));
  float     * FloatTmp = (float *)allocate(sizeof(float));

  for (i = 0; i < bookSize; i ++)
  {
    for (j = i + 1; j < bookSize; j ++)
    {
      idx = (2 * bookSize - i - 1) * i / 2 + (j - i) - 1;
      Pairs[idx].index1 = i;
      Pairs[idx].index2 = j;

      EuclideanDistII(TS->Book[i].vector, 
                      TS->Book[j].vector, 
                      VectorSize(TS),
                      FloatTmp);

      Pairs[idx].weight = *FloatTmp;

    }
  }

  PInfo->Pairs = Pairs;
  PInfo->PairSize = bookSize * (bookSize - 1) / 2;

  deallocate(FloatTmp);
  return PInfo;
}

/*-------------------------------------------------------------------*/
/* To produce pairs (edges) of the complete graph
 * Pair<weight, index1, index2>                                      
 */ 

PairInfo * OrganizePairsFromVectorSet(float  ** Vectors, 
                                      int       vectorSize, 
                                      int       bookSize)
{
  int         i, j, idx;

  Pair      * Pairs = (Pair*)allocate(sizeof(Pair) * bookSize * (bookSize - 1) / 2);
  PairInfo  * PInfo = (PairInfo*)allocate(sizeof(PairInfo));
  float     * FloatTmp = (float *)allocate(sizeof(float));

  if (Pairs == NULL) ;//error

  for (i = 0; i < bookSize; i ++){
    for (j = i + 1; j < bookSize; j ++){
      idx = (2 * bookSize - i - 1) * i / 2 + (j - i) - 1;
      Pairs[idx].index1 = i;
      Pairs[idx].index2 = j;

      EuclideanDistFF(Vectors[i], 
                      Vectors[j], 
                      vectorSize,
                      FloatTmp);
      Pairs[idx].weight = *FloatTmp;
    }
  }

  PInfo->Pairs = Pairs;
  PInfo->PairSize = bookSize * (bookSize - 1) / 2;

  deallocate(FloatTmp);
  return PInfo;
}

/*-------------------------------------------------------------------*/

void SwapPairs(Pair *p1, Pair *p2){
  Pair h;

  h.index1   = p1->index1; 
  h.index2   = p1->index2; 
  h.weight   = p1->weight; 

  p1->index1 = p2->index1; 
  p1->index2 = p2->index2; 
  p1->weight = p2->weight; 

  p2->index1 = h.index1;            
  p2->index2 = h.index2;            
  p2->weight = h.weight;  
  
}

/*-------------------------------------------------------------------*/

void InsertSortPairs(Pair s[], int lo, int hi){
  int i, j;

  for (i = lo; i < hi; i ++)
    for (j = i + 1; j <= hi; j ++){
      if (s[i].weight > s[j].weight){
        SwapPairs(&s[i], &s[j]);
      }
    }
}

/*-------------------------------------------------------------------*/

void QuickSortPairs (Pair s[], int lo, int hi){
  //lo is the lower index, hi is the upper index
  //of the region of array a that is to be sorted    
  int i = lo, j = hi;
  Pair x;    

  if (hi - lo < 5){
    InsertSortPairs(s, lo, hi);
    return;
  }
  
  x.index1 = s[(lo+hi) / 2].index1;    
  x.index2 = s[(lo+hi) / 2].index2;    
  x.weight = s[(lo+hi) / 2].weight;    

  //partition    
  do{            
    while (s[i].weight < x.weight) i++;         
    while (s[j].weight > x.weight) j--;        
    if (i <= j){             
      SwapPairs(&s[i], &s[j]);
      i++; 
      j--;        
    }    
  } while (i<=j);  
  
  //recursion    
  if (lo <j ) QuickSortPairs(s, lo, j);    
  if (i < hi) QuickSortPairs(s, i, hi);
}


/*-------------------------------------------------------------------*/
/* 
 *  PairResult: return pairs of MST and corresponding indices of the 
 *  edge in PairInfo, with last extra element denoting valid length 
 *  of pairs. The indices of the edge in PairInfo are reserved for 
 *  constructing kMST in which previous MST edges must be removed.
 *
 *  If this function is only used to construct a MST, the information
 *  of the indices (i.e. the third element in PairResult) is useless.
 */

int * MST_Kruskal(int bookSize, PairInfo *PInfo)
{
  int   i, j, rep1, rep2;
  int   num_fill = 0;

  // First two elements are edge endpoint indices, the third elements
  // is the index of the edge in PairInfo->Pair, the last extra for length.
  int * PairResult = (int *)allocate(sizeof(int) * (bookSize - 1) * 3 + 1);

  int * Representative = (int *)allocate(sizeof(int) * bookSize);

  //initialize
  for (i = 0; i < bookSize; i ++) {
    Representative[i] = -1;
  }

  // sort the pairs by weights
  QuickSortPairs(PInfo->Pairs, 0, PInfo->PairSize - 1);  

  //Kruskal process
  for (i = 0; i < PInfo->PairSize; i ++)
  {
    rep1 = Representative[PInfo->Pairs[i].index1];
    rep2 = Representative[PInfo->Pairs[i].index2];
   
    if (rep1 != rep2 || (rep1 < 0 && rep2 < 0))
    {    

      PairResult[num_fill * 3] = PInfo->Pairs[i].index1;
      PairResult[num_fill * 3 + 1] = PInfo->Pairs[i].index2;
      PairResult[num_fill * 3 + 2] = i;
        
      if (rep1 < 0 && rep2 < 0){
        Representative[PInfo->Pairs[i].index1] =PInfo->Pairs[i].index1;
        Representative[PInfo->Pairs[i].index2] =PInfo->Pairs[i].index1;
      }else if (rep1 < 0 && rep2 >= 0){
        Representative[PInfo->Pairs[i].index1] = rep2;
      }else if (rep1 >= 0 && rep2 < 0){
        Representative[PInfo->Pairs[i].index2] = rep1;
      }else {
        for (j = 0; j < bookSize; j ++){     // combine two subtrees 
          if (Representative[j] == rep2){
            Representative[j] = rep1;
          }
        }
      }

      num_fill ++;  

      if(num_fill>= bookSize-1) break;
    }
  }
  PairResult[(bookSize - 1) * 3] = num_fill;  //last element denotes valid length
  deallocate(Representative);
  return PairResult;
}


/*-------------------------------------------------------------------*/
/* 
  Convert MST pairs into DGR Graph                                  
*/

Graph * MSTGraph(TRAININGSET *TS, int * PairResult)
{
  int         * graphdata, maxnum, * graphresult;
  int           i, j, idx1, idx2, dim, bookSize = BookSize(TS), tmp = 0;
  Graph       * g;
  GraphVector * v;

  int           num_fill = PairResult[(bookSize - 1) * 3];

  // construct GRD graph
  graphdata = (int *)allocate(sizeof(int) * bookSize * bookSize);
  for (i = 0; i < bookSize; i ++){
    graphdata[i * bookSize + bookSize - 1] = 0;
  }

  maxnum = 0;

  for (i = 0; i < num_fill; i ++){
    idx1 = PairResult[i * 3];
    idx2 = PairResult[i * 3 + 1];

    graphdata[idx1 * bookSize + bookSize - 1] ++;
    
    if (graphdata[idx1 * bookSize + bookSize - 1] > maxnum){
      maxnum = graphdata[idx1 * bookSize + bookSize - 1];
    }

    tmp = graphdata[idx1 * bookSize + bookSize - 1];
    graphdata[idx1 * bookSize + tmp - 1] = idx2;
  }

  for (i = 0; i < bookSize; i ++){
    tmp = graphdata[i * bookSize + bookSize - 1];
    for (j = tmp + 1; j <= maxnum; j ++){
        graphdata[i * bookSize + j - 1] = i;
    }
  }

  dim         = VectorSize(TS);
  graphresult = (int *)allocate(sizeof(int) * bookSize * (maxnum+ dim));
  
  for (i = 0; i < bookSize; i ++){
    graphresult[i * (maxnum+ dim)] = TS->Book[i].vector[0];
    graphresult[i * (maxnum+ dim) + 1] = TS->Book[i].vector[1];

    for (j = dim; j < maxnum+ dim; j ++){
      graphresult[i * (maxnum+ dim) + j] = graphdata[i * bookSize + j - dim];
    }
  }

  g = AllocateMemoryForGraph(bookSize, maxnum, dim);  // This function has a problem: it allocates memory
  for (i = 0; i < bookSize; i ++){                    // for vectors, but AllocateMemoryForGraphVector
                                                      // allocates again.
    v = AllocateMemoryForGraphVector(g, i);
    for (j = 0; j < dim; j ++){
      v->data[j] = graphresult[i * (maxnum+ dim) + j];
    }

    for (j = 0; j < maxnum; j ++){
      v->kindices[j] = graphresult[i * (maxnum+ dim) + j + dim];
    }

    v->index = i;
    g->vectors[i] = v;
  }

  g->nvec = bookSize;
  g->k = maxnum;
  g->dim = dim;
 
  deallocate(graphdata);
  deallocate(graphresult);

  return g;
}

/*-------------------------------------------------------------------*/
/*
  This function is to construct MST from Training set 
  and return corresponding Pairs
*/

int * MST_KruskalFromTrainingSetReturnPair(TRAININGSET *TS){
  //FILE    * fp;
  int     * PairResult;

  // organize the pairs
  PairInfo * PInfo = OrganizePairsFromTrainingSet(TS);

  // construct MST
  PairResult = MST_Kruskal(BookSize(TS), PInfo);

  /*
  fp = fopen("e:\\ClassicalMST.txt","w");
  for (i = 0; i < BookSize(TS)-1; i ++){
        fprintf(fp, "%d ", PairResult[i * 3]);
        fprintf(fp, "%d ", PairResult[i * 3+1]);

        fprintf(fp, "\n");
  }
  fclose(fp);*/

  deallocate(PInfo->Pairs);
  deallocate(PInfo);
  
  return PairResult;
}

/*-------------------------------------------------------------------*/
/*
  This function is to construct MST from Pairs (see PairInfo structure) 
  and return corresponding Pairs
*/

int * MST_KruskalFromPairsReturnPair(TRAININGSET *TS, PairInfo *PInfo){
  int * PairResult;

  // construct MST
  PairResult = MST_Kruskal(BookSize(TS), PInfo);

  return PairResult;
}

/*-------------------------------------------------------------------*/
/*
  This function is to construct MST from Vectors
  int vectorSize: Dimension
  int bookSize: Number of vectors
  and return corresponding Pairs
*/

int * MST_KruskalFromVectorSetReturnPair(float ** Vectors, int vectorSize, int bookSize){
  int * PairResult;

  // organize the pairs
  PairInfo * PInfo = OrganizePairsFromVectorSet(Vectors, vectorSize, bookSize);

  // construct MST
  PairResult = MST_Kruskal(bookSize, PInfo);

  deallocate(PInfo->Pairs);
  deallocate(PInfo);

  return PairResult;
}

/*-------------------------------------------------------------------*/
/*
  This function is to construct MST from Training set 
  and return corresponding GRAPH
*/

Graph * MST_KruskalFromTrainingSet(TRAININGSET *TS)
{
  int * PairResult = MST_KruskalFromTrainingSetReturnPair(TS); 
  Graph * g;

  // graph
  g = MSTGraph(BookSize(TS), PairResult);

  deallocate(PairResult);

  return g;
}


/*-------------------------------------------------------------------*/
/*
  This function is to construct MST from Pairs (see PairInfo structure) 
  and return corresponding GRAPH
*/

Graph * MST_KruskalFromPairs(TRAININGSET *TS, PairInfo *PInfo)
{
  int * PairResult = MST_KruskalFromPairsReturnPair(TS, PInfo);
  Graph * g;

  // graph
  g = MSTGraph(BookSize(TS), PairResult);

  deallocate(PairResult);

  return g;

}
