/*-------------------------------------------------------------------*/
/* PNNGRAPH.C          Pasi Frnti & Olli Virmajoki                  */
/*                                                                   */
/*  Graph PNN method with PDS & MPS & Lazy                           */
/*  Starts from initial solution                                     */
/*	 Reduces cluster size to FinalSize                           */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GRAPH"
#define VersionNumber  "Version 0.32"
#define LastUpdated    "05.08.2009" /* MM */

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <values.h>

#include "bintree.h"
#include "heap.h"
#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnngraph.h"
#include "pnn.h"
#include "split.h"
#include "owntypes.h"
#include "sa.h"


/*---------------------------  M i s c	------------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong ksum=0LL;
static llong ksumtotal=0LL;
static llong ndistcalcs=0LL;
static llong ndistcalcstotal=0LL;
static llong nnccalcs=0LL;
static llong nnccalcstotal=0LL;
static llong ntau=0LL;
static llong ntautotal=0LL;
static llong nterm=0LL;
static llong npre=0LL;
static llong npretotal=0LL;
static llong niteration=0LL;
STATCODE(
static llong steps=0LL;
static llong steps1=0LL;
static llong steps2=0LL;
static llong steps3=0LL;
static llong steps4=0LL;
static llong steps5=0LL;
static llong dists=0LL;
static llong d1=0LL;
static llong d2=0LL;
static llong d3=0LL;
static llong d4=0LL;
static llong d5=0LL;)

static void FindNearestClusterMPS(CODEBOOK* CB, DATA* D, int a, KNNDATA* KNN, int K, int Mlf);
static void AddToCC(TRAININGSET*  TS,
                    PARTITIONING* P,
                    int           TSvec,
                    int           Pindex);

/* ================================================================= */

static DISTELEM* DM_elem(KDATA* KD, int i)
{
  return &(KD->DM[i][0]) ;
}

/* ----------------------------------------------------------------- */

static int DM_from(KDATA* KD, int i)
{
  return( KD->DM[i][0].from );
}

/* ----------------------------------------------------------------- */

static int DM_to(KDATA* KD, int i)
{
  return( KD->DM[i][0].to );
}

/* ----------------------------------------------------------------- */

static int DM_heapindex(KDATA* KD, int i)
{
  return( KD->DM[i][0].heapindex );
}

/* ----------------------------------------------------------------- */

static int* DM_heapindexaddr(KDATA* KD, int i)
{
  return( &(KD->DM[i][0].heapindex) );
}

/* ----------------------------------------------------------------- */

static int DM_dist(KDATA* KD, int i)
{
  return( KD->DM[i][0].dist );
}

/* ----------------------------------------------------------------- */

static void DM_distset(KDATA* KD, int i, llong distp)
{
  KD->DM[i][0].dist = distp;
}

/* ----------------------------------------------------------------- */

static void DM_toset(KDATA* KD, int i, int j)
{
  KD->DM[i][0].to = j;
}


/*============================  P R I N T  ===============================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


static void PrintProgress(TRAININGSET*   TS,
                          CODEBOOK*      CB,
                          PARTITIONING*  P,
                          int            Quiet,
                          long           watch)
{
  double Error=0.0;
  PARTITIONING PT;

  if(Quiet<2) return;
  ksumtotal += ksum;
  ndistcalcstotal += ndistcalcs;
  nnccalcstotal += nnccalcs;
  ntautotal += ntau;
  npretotal += npre;
  niteration += 1;

  switch(Quiet)
  {
        case 2:
        if(BookSize(CB) < 256)
        {         
        CreateNewPartitioning(&PT, TS, BookSize(CB));
        CopyPartitioning(P, &PT);
        GenerateOptimalPartitioning(TS, CB, &PT);
        Error = AverageErrorForSolution(TS, CB, &PT, MSE);
        Error = PrintableError(Error, CB);
        FreePartitioning(&PT);
        printf("\n%3i: %9.4f\ntau: %7.2f term: %7.2f\ndist calcs/search: %7.2f dim/dist calcs: %7.2f dim/seaerch: %7.2f \nksum: %7.0f ndistcalcs: %7.0f searches: %7.0f",
        BookSize(CB), Error,
        (double)ntau, (double)nterm,
        nnccalcs != 0LL ? (double)ndistcalcs/nnccalcs : 0.0,
        ndistcalcs != 0LL ? (double)ksum/ndistcalcs : 0.0,
        nnccalcs != 0LL ? (double)ksum/nnccalcs : 0.0,
        (double)ksum,(double)ndistcalcs,(double)nnccalcs);
        }
        break;
    case 3:  
        printf("%3i: %9.4f ",  BookSize(CB), Error); PrintTime(watch); break;
    default: 
        printf("Book size = %3i   Error=%9.4f  Time=", BookSize(CB), Error);
             PrintTime(watch); break;
  }
  fflush(stdout);
  ksum=0LL;
  ndistcalcs=0LL;
  nnccalcs=0LL;
  ntau=0LL;
  npre=0LL;
  nterm=0LL;
}


/*====================================================================*/


static llong MergeDistortion(CODEBOOK* CB, int p1, int p2)
{
  /* Original "inline" */
  int i=0, Vsize;
  llong   dist=0LL, diff, dist_limit=MAXLLONG;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
	 /* Precondition: Vsize >= 1 */
     /* if dist >= dist_limit, returns at least factor * dist_limit */
  do
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
           ksum++; /* Statistical information */
           )
	} while(dist < dist_limit && ++i < Vsize);
  dist=round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )

  return( dist);
}


/*-------------------------------------------------------------------*/


static llong MergeDistortionPdsSIM(CODEBOOK* CB, int p1, int p2,
								llong dist_limit)
{
  /* Variant Simple */
  int i=0, Vsize;
  llong   dist=0LL, diff, distortion=0LL;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  /* Precondition: Vsize >= 1 */
  /* if distortion >= dist_limit, returns at least dist_limit */
  do	
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
           ksum++; /* Statistical information */
           )
	   distortion=round( factor * (double)dist);
	} while( distortion < dist_limit && ++i < Vsize);
    STATCODE
        (
        ndistcalcs++; /* Statistical information */
        )
  return( distortion);
}


static llong MergeDistortionPds(CODEBOOK* CB, int p1, int p2,
								  llong maxdist)
{
  /* Variant Optimized */
  int i=0, Vsize;
  llong   dist=0LL, diff, distortion=0LL, dist_limit;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  dist_limit = roundpds((double)maxdist/factor);
  /* Precondition: Vsize >= 1 */
  /* if dist >= dist_limit, returns at least maxdist */
  do
	 {
		diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
		dist += diff * diff;
        STATCODE
            (
    		ksum++; /* Statistical information */
            )
	 } while(dist < dist_limit && ++i < Vsize);
  distortion=round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )
  return( distortion);
}


/*-------------------------------------------------------------------*/


static void MarkClustersForRecalculation(CODEBOOK* CB,
                                         KNNDATA* KNN, int K,
                                         int a, int b)
{
  int i,j;
  j=0;
    for( i=0; i<BookSize(CB); i++ )
        {
            STATCODE(steps2++;)
            if(KNN->NN[i][j].Vector==a || KNN->NN[i][j].Vector==b)
            KNN->NN[i][j].Recalculate = YES;
	    }
}


/*-------------------------------------------------------------------*/


static void SaveNeighbor(KNNDATA* KNN, int K, llong Dist, int a, int b)
{
    int i,j;
    for( i=0; i<K; i++)
    {
        if(Dist < KNN->NN[a][i].VectorDist)
        {
            for( j=K-1; j>i; j--)
            {
                KNN->NN[a][j].VectorDist = KNN->NN[a][j-1].VectorDist;
                KNN->NN[a][j].Vector = KNN->NN[a][j-1].Vector;
                KNN->NN[a][j].Recalculate = KNN->NN[a][j-1].Recalculate;
            }
            KNN->NN[a][i].VectorDist = Dist;
            KNN->NN[a][i].Vector = b;
            KNN->NN[a][i].Recalculate = NO;
            /*break; do not work*/
            return;
        }
    }
}


/*-------------------------------------------------------------------*/


static void CalculateAllDistancesGraphFast(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
 KDATA* KD, HEAP* H, int Heap, int Quiet, long watch, int Pds, int Mps, int Mlf)
/* Brute Force / fast */
{
  int  i,j=0;
  int L=K-1;
  llong Dist=MAXLLONG;

  if(Quiet>=3) printf("Calculating distance matrix:\n"); fflush(stdout);

  if(Mps) /*MPS*/
  {
  for( i=0; i<BookSize(CB); i++)
    {
	    FindNearestClusterMPS(CB, D, i, KNN, K, Mlf);
        if(Heap)
        {
        DM_distset(KD, i, KNN->NN[i][0].VectorDist);
        DM_toset(KD, i, KNN->NN[i][0].Vector);
        Heap_insert(&(KD->H), DM_elem(KD, i), NULL, DM_heapindexaddr(KD, i));
        }
        if(ExitRequested()) ExitProcessing(-2);
    }
  }
  else /*Brute Force / fast */
  {
  for( i=0; i<BookSize(CB)-1; i++)
  {
	 for(j=i+1; j<BookSize(CB);j++)
     {
		  if(Pds)
		  {
             Dist=MergeDistortionPds(CB,i,j,KNN->NN[i][L].VectorDist);
		  }
		  else
		  {
			 Dist=MergeDistortion(CB,i,j);
		  }
		  if(Dist<KNN->NN[i][L].VectorDist)
		  {
             SaveNeighbor(KNN, K, Dist, i, j);
		  }

     if(ExitRequested()) ExitProcessing(-2);
     }
    if(Heap)
    {
        DM_distset(KD, i, KNN->NN[i][0].VectorDist);
        DM_toset(KD, i, KNN->NN[i][0].Vector);
        Heap_insert(&(KD->H), DM_elem(KD, i), NULL, DM_heapindexaddr(KD, i));
    }
  }
  if(Heap)
  {
     i=BookSize(CB)-1;
     DM_distset(KD, i, MAXLLONG/10);
     DM_toset(KD, i, i);
     Heap_insert(&(KD->H), DM_elem(KD, i), NULL, DM_heapindexaddr(KD, i));
  }
  }
  STATCODE
      (
      nnccalcs=BookSize(CB);
      )
    if(Quiet>=3) { printf("\r                    \r"); PrintTime(watch); }
}


/*-------------------------------------------------------------------*/


static void CalculateAllDistancesGraph(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
    KDATA* KD, HEAP* H, int Heap, int Quiet, long watch, int Pds)
/* Brute Force */
{
  int  i,j;
  int L=K-1;
  llong Dist=MAXLLONG;

  if(Quiet>=3) printf("Calculating distance matrix:\n"); fflush(stdout);

    for( i=0; i<BookSize(CB); i++)
    {
     for( j=0; j<BookSize(CB); j++)
     {
        if(i==j)continue;
		  if(Pds)
		  {
             Dist=MergeDistortionPds(CB,i,j,KNN->NN[i][L].VectorDist);
		  }
		  else
		  {
			 Dist=MergeDistortion(CB,i,j);
		  }
		  if(Dist<KNN->NN[i][L].VectorDist)
		  {
             SaveNeighbor(KNN, K, Dist, i, j);
		  }
     
     if(ExitRequested()) ExitProcessing(-2);
     }
     if(Heap)
     {
        DM_distset(KD, i, KNN->NN[i][0].VectorDist);
        DM_toset(KD, i, KNN->NN[i][0].Vector);
        Heap_insert(&(KD->H), DM_elem(KD, i), NULL, DM_heapindexaddr(KD, i));
     }
    }
  STATCODE
      (
      nnccalcs=BookSize(CB);
      )
    if(Quiet>=3) { printf("\r                    \r"); PrintTime(watch); }
}


/*-------------------------------------------------------------------*/


static void* Heap_getelemdata(HEAP* H, int heapindex, void* info)
{
    void* elem;
    elem = H->data[heapindex];
    return(elem);
}


/*-------------------------------------------------------------------*/


static void FindMergedPairGraph(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
 KDATA* KD, HEAP* H, int Heap, int* a, int* b, int Pds, int Mps)
{
  llong MinDist=MAXLLONG, Dist;
  int i;
  int nrec=0;
  DISTELEM* cell;

  if(Heap)
  {
     cell = Heap_getelemdata(&(KD->H), 1, NULL);
     *a = cell->from;
     *b = cell->to;
     STATCODE(steps1++;)
  }

  else /*Linear search and lazy*/
  {
    for( i=0; i<BookSize(CB); i++ )
    {
     STATCODE(steps1++;)
     if( KNN->NN[i][0].VectorDist < MinDist )
        if(KNN->NN[i][0].Recalculate && i!=KNN->NN[i][0].Vector) /*Lazy*/
        {
           Dist=MergeDistortion(CB,i,KNN->NN[i][0].Vector);
           KNN->NN[i][0].VectorDist = Dist;
           KNN->NN[i][0].Recalculate = NO;
           nrec++;
           if(KNN->NN[i][0].VectorDist < MinDist)
           {
              (*a) = i;
              (*b) = KNN->NN[i][0].Vector;
              MinDist = KNN->NN[i][0].VectorDist;
           }
        }
        else
        {
           (*a) = i;
           (*b) = KNN->NN[i][0].Vector;
           MinDist = KNN->NN[i][0].VectorDist;
        }
    }
  }

  if( (*a)>(*b) ) swap((*a),(*b));

  /*printf("\nFound %d %d %f BS= %d",*a,*b,(float)KNN->NN[*a][0].VectorDist, BookSize(CB));*/
  STATCODE
      (
      ntau=nrec;
      nnccalcs=nrec;
      )
}


/*-------------------------------------------------------------------*/


static void MergeVectorsGraph(TRAININGSET*  TS,
                  CODEBOOK*     CB,
                  PARTITIONING* P,
                  KNNDATA*      KNN,
                  int           K,
                  KDATA*        KD,
                  HEAP*         H,
                  int           Heap,
                  int           a,
                  int           b,
                  int			Lazy)
{
  NEARESTTYPE APU[K];
  llong Dist;
  int i,j,flag;
  int L=K-1;
  int nrec=0;
  int last = BookSize(CB)-1;

  for( j=0; j<K; j++)
  {
     APU[j].VectorDist=KNN->NN[a][j].VectorDist;
     APU[j].Vector=KNN->NN[a][j].Vector;
     APU[j].Recalculate=KNN->NN[a][j].Recalculate;
     KNN->NN[a][j].VectorDist = MAXLLONG/10;
     KNN->NN[a][j].Vector=a;
     KNN->NN[a][j].Recalculate=NO;
     STATCODE(steps2++;)
  }
  if(Lazy)MarkClustersForRecalculation(CB, KNN, K, a, b);
  JoinPartitions(TS, P, a, b);
  PartitionCentroid(P, a, &Node(CB, a));
  for( j=0; j<K; j++)
  {
     if(APU[j].Vector == b || APU[j].Vector == a)continue;
     flag=0;
     for(i=0; i<K; i++)
     {
         if(APU[j].Vector==KNN->NN[a][i].Vector)
             flag=1;
     STATCODE(steps2++;)
     }
     if(flag)
     {
         /*printf("\nNot OK!");*/
         continue;
     }
     STATCODE(d2++;)
     Dist = MergeDistortion(CB, a, APU[j].Vector);
     nrec++;
     if(Dist<KNN->NN[a][L].VectorDist)
     {
     		SaveNeighbor(KNN, K, Dist, a, APU[j].Vector);
     }
  }
  for( j=0; j<K; j++)
  {
     if(KNN->NN[b][j].Vector == a || KNN->NN[b][j].Vector == b)continue;
     flag=0;
     for(i=0; i<K; i++)
     {
         if(KNN->NN[b][j].Vector==KNN->NN[a][i].Vector)
             flag=1;
     STATCODE(steps2++;)
     }
     if(flag)continue;
     STATCODE(d2++;)
     Dist = MergeDistortion(CB, a, KNN->NN[b][j].Vector);
     nrec++;
     if(Dist<KNN->NN[a][L].VectorDist)
     {
     		SaveNeighbor(KNN, K, Dist, a, KNN->NN[b][j].Vector);
     }
  }
  if(Heap)
  {
     DM_distset(KD, a, KNN->NN[a][0].VectorDist);
     DM_toset(KD, a, KNN->NN[a][0].Vector);
     Heap_update(&(KD->H),DM_heapindex(KD, a),NULL);
     STATCODE(steps2=steps2+round(log10(BookSize(CB))/log10(2.));)
  }
  for( i=0; i<BookSize(CB); i++)
      for( j=0; j<K; j++)
        {
            STATCODE(steps3++;)
            if(KNN->NN[i][j].Vector==b)
            {
            KNN->NN[i][j].Vector=a;
            if(Heap && j==0) DM_toset(KD, i, KNN->NN[i][0].Vector);
            }
            if( KNN->NN[i][j].Vector==last )
            {
            KNN->NN[i][j].Vector = b;
            if(Heap && j==0) DM_toset(KD, i, KNN->NN[i][0].Vector);
            }
        }
  /*RemoveLastVector*/
  if( b != last )
  {
    CopyNode(&Node(CB,last), &Node(CB,b), VectorSize(CB));
    for( j=0; j<K; j++)
    {
    STATCODE(steps4++;)
    KNN->NN[b][j].VectorDist = KNN->NN[last][j].VectorDist;
    KNN->NN[b][j].Vector = KNN->NN[last][j].Vector;
    KNN->NN[b][j].Recalculate = KNN->NN[last][j].Recalculate;
    }
    if(Heap)
    {
       DM_distset(KD, b, KNN->NN[b][0].VectorDist);
       DM_toset(KD, b, KNN->NN[b][0].Vector);
       Heap_update(&(KD->H),DM_heapindex(KD, b),NULL);
       STATCODE(steps4=steps4+round(log10(BookSize(CB))/log10(2.));)
    }
  }
  if(Heap)
  {
     Heap_remove(&(KD->H), DM_heapindex(KD, last), NULL);
     STATCODE(steps4=steps4+round(log10(BookSize(CB))/log10(2.));)
  }
  BookSize(CB)--;
  STATCODE
      (
      ntau=nrec;
      nnccalcs=nrec;
      )
}


/*-------------------------------------------------------------------*/


static void RecalculateDistancesGraph(CODEBOOK* CB, KNNDATA* KNN, int K,
 KDATA* KD, HEAP* H, int Heap, int aex, int Quiet, int Pds, int Lazy)
{
  int  i,j;
  int  nrec=0;
  llong Dist;

  if(Lazy)return;
  j=0;
  for( i=0; i<BookSize(CB); i++ )
    {
    STATCODE(steps5++;)
    if( KNN->NN[i][j].Vector==aex && (i != KNN->NN[i][j].Vector))
       {
        STATCODE(d5++;)
        Dist=MergeDistortion(CB, i, KNN->NN[i][j].Vector);
	     KNN->NN[i][j].VectorDist=Dist;
        nrec++;
        if(Heap)
        {
        DM_distset(KD, i, KNN->NN[i][0].VectorDist);
        Heap_update(&(KD->H), DM_heapindex(KD, i), NULL);
        STATCODE(steps5=steps5+round(log10(BookSize(CB))/log10(2.));)
        }
       }
       KNN->NN[i][j].Recalculate = NO;
    }
  STATCODE
      (
      nnccalcs+=nrec;
      ntau+=nrec;
      )
  if(Quiet>=4) printf("Recalculated: %i / %i.\n", nrec, BookSize(CB));
}


/* ========================  MPS  ================================== */


static int cmpCodeSum(const void* e1, const void* e2, const void* info)
{
  return( ((int*)info)[*((int*)e1)] < ((int*)info)[*((int*)e2)] ? 1 : 0 );
}


/*-------------------------------------------------------------------*/


static int VectorSum(int* vector, int dim)
{
  int sum = 0;
  int k;

  for( k = 0; k < dim; k++ )
    {
    sum += vector[k];
    }
  return( sum );
}


/*-------------------------------------------------------------------*/


static void CalculateCodevectorSums(CODEBOOK* CB, DATA* D)
{
  int i;

  for( i = 0; i < BookSize(CB); i++ )
    {
    D->codesum[i] = VectorSum(Vector(CB, i), VectorSize(CB));
    }
}


/*-------------------------------------------------------------------*/


static llong SMD(int sum1, int sum2)
{
  return( (llong)(sum1 - sum2) * (llong)(sum1 - sum2) );
}


/*-------------------------------------------------------------------*/


static void MPSsearch(int		  a,
                     CODEBOOK*    CB,
                     llong*	      dist,
                     KNNDATA*	  KNN,
                     int		  K,
                     int          Mlf,
                     int          SMDindex,
                     int          start,
                     int          end,
                     int          sum[],
                     int          order[])
{
  int up=1, down=1;
  int m=SMDindex, n=SMDindex;
  int L=K-1;
  int count=0;
  llong Dist;
  STATCODE
      (
      int termup=0;
      int termdown=0;
      )

  while(up || down)
  {
      if(down)
      {
          m++;
          if(m > end)
             down=0;
          else
          {
              STATCODE
                  (
                  npre += 1;
                  )
              if(
		        (((double)VectorFreq(CB,a) * (double)VectorFreq(CB,order[m]) /
                ((double)VectorFreq(CB,a) + (double)VectorFreq(CB,order[m]))) * 
		        SMD(sum[a], sum[order[m]]))
		         > VectorSize(CB) * (*dist)
	        	)
              {
                if(VectorFreq(CB,order[m])==1)
                {
                      down=0;
                      STATCODE
                          (
                          termdown=1;
                          )
                }
              }
              else
              {
                 Dist=MergeDistortionPds(CB, a, order[m], KNN->NN[a][L].VectorDist);
                   if(Dist < KNN->NN[a][L].VectorDist)
                   {
                   SaveNeighbor(KNN, K, Dist, a, order[m]);
                   if(Dist < *dist)
                   {
			          *dist=Dist;
			          /*b=order[m];*/
                   }
                   }
                 if(++count>=K+Mlf)return;
              }
          }
      }
      if(up)
      {
          n--;
          if(n < start)
            up=0;
          else
          {
              STATCODE
                  (
                  npre += 1;
                  )
              if(
		        (((double)VectorFreq(CB,a) * (double)VectorFreq(CB,order[n]) /
                ((double)VectorFreq(CB,a) + (double)VectorFreq(CB,order[n]))) *
		        SMD(sum[a], sum[order[n]]))
		         > VectorSize(CB) * (*dist)
	        	)
              {
                 if(VectorFreq(CB,order[n])==1)
                 {
                    up=0;
                    STATCODE
                        (
                        termup=1;
                        )
                 }
              }
              else
              {
                 Dist=MergeDistortionPds(CB, a, order[n], KNN->NN[a][L].VectorDist);
                    if(Dist < KNN->NN[a][L].VectorDist)
                    {
                    SaveNeighbor(KNN, K, Dist, a, order[n]);
                    if(Dist < *dist)
                    {
			           *dist=Dist;
			           /*b=order[n];*/
                    }
                    }
                 if(++count>=K+Mlf)return;
              }
          }
      }
  }
  /*printf("\ncount = %d",count);*/
  STATCODE
      (
      if(termup||termdown)
      nterm+=1;
      )

}


/*-------------------------------------------------------------------*/


static void FindNearestClusterMPS(CODEBOOK* CB, DATA* D, int a, 
                                  KNNDATA* KNN, int K, int Mlf)
{
  int SMDindex=0;
  llong distH;
  llong* dist=&distH;

  SMDindex = D->mynumber[a];
  *dist = MAXLLONG/VectorSize(CB);
  MPSsearch(a, CB, dist, KNN, K, Mlf,
	  SMDindex, 0, BookSize(CB)-1, D->codesum, D->codeorder);
}


/* ================================================================= */


static void PnnGraphWriteGraph(char *fname, CODEBOOK* CB, KNNDATA* KNN, int K, int FinalSize)
{
  int i, j, currcb = 0, *lut;
  Graph *g_out;
  GraphVector *v;

  lut = (int *) allocate(sizeof(int)*BookSize(CB));
  
  g_out = AllocateMemoryForGraph(FinalSize, K, VectorSize(CB));


  for ( i = 0; i < BookSize(CB); i++) {
    if (VectorFreq(CB,i) != 0) {
      lut[i] = currcb;
      printf("%d %d\n", i, currcb);
      currcb++;
    } else {
      lut[i] = -1;
    }
  }    

  for ( i = 0; i < BookSize(CB); i++ ) {
    if (VectorFreq(CB,i) != 0) {
      v = AllocateMemoryForGraphVector(g_out, lut[i]); 
      for ( j = 0; j < VectorSize(CB); j++) {
	GraphPutVectorElement(v, j, VectorScalar(CB,i,j));
      }    
      for ( j = 0; j < K; j++ ) {
	GraphPutNN(g_out, v, lut[KNN->NN[i][j].Vector], j); 	  
      }
      printf("%p  %p\n", v, g_out->vectors[lut[i]]);
      g_out->vectors[lut[i]] = v;
    }
  }
  GraphWrite(fname, g_out, YES);
}


/* ================================================================= */


static Graph* PnnGraphFormGraph(CODEBOOK* CB, KNNDATA* KNN, int K,   	                             int FinalSize)
{
  int i, j, currcb = 0, *lut;
  GraphVector *v;
  Graph *g;

  lut = (int *) allocate(sizeof(int)*BookSize(CB));

  g = AllocateMemoryForGraph(FinalSize, K, VectorSize(CB));


  for ( i = 0; i < BookSize(CB); i++) {
    if (VectorFreq(CB,i) != 0) {
      lut[i] = currcb;
      printf("%d %d\n", i, currcb);
      currcb++;
    } else {
      lut[i] = -1;
    }
  }

  for ( i = 0; i < BookSize(CB); i++ ) {
    if (VectorFreq(CB,i) != 0) {
      v = AllocateMemoryForGraphVector(g, lut[i]);
      for ( j = 0; j < VectorSize(CB); j++) {
        GraphPutVectorElement(v, j, VectorScalar(CB,i,j));
      }
      for ( j = 0; j < K; j++ ) {
        GraphPutNN(g, v, lut[KNN->NN[i][j].Vector], j);
      }
      printf("%p  %p\n", v, g->vectors[lut[i]]);
      g->vectors[lut[i]] = v;
    }
  }
  return g;
}



static void CreateDataStructures(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
KDATA* KD, HEAP* H, int Heap, int Mps)
{
   int i,j;
   if(Mps)
   {
      D->codesum = allocate(BookSize(CB) * sizeof(int));
	  D->codeorder = allocate(BookSize(CB) * sizeof(int));
	  D->mynumber = allocate(BookSize(CB) * sizeof(int));
   }
   KNN->NN = (NEARESTTYPE**)allocate(BookSize(CB) * sizeof(NEARESTTYPE*));
   KNN->head = (VERTEX**)allocate(BookSize(CB)*sizeof(VERTEX*));
      for( i = 0; i < BookSize(CB); i++ )
         {
         KNN->NN[i] = (NEARESTTYPE*)allocate(K * sizeof(NEARESTTYPE));
         KNN->head[i] = NULL;
         for( j = 0; j < K; j++ )
         {
            KNN->NN[i][j].VectorDist = MAXLLONG/10;
            KNN->NN[i][j].Vector = i;
            KNN->NN[i][j].Recalculate = NO;
         }
         }
   if(Heap)
   {
      Heap_init(&(KD->H), BookSize(CB), CmpDistKurita );
      KD->DM = (DISTELEM**)allocate(BookSize(CB) * sizeof(DISTELEM*));
      for(i=0; i<BookSize(CB); i++)
      {
         KD->DM[i] = (DISTELEM*)allocate(sizeof(DISTELEM));
         KD->DM[i][0].from = i;
         KD->DM[i][0].to = i;
         KD->DM[i][0].heapindex = 0;
         KD->DM[i][0].dist = 0LL;
      }
   }
}


static void InitializeData(CODEBOOK* CB, DATA* D, int Mps)
{
    int i, j;
    if(Mps)
    {
	   for(i=0; i<BookSize(CB); i++)
       {
		    D->codesum[i] = 0;
		    D->codeorder[i] = i;
       }
       CalculateCodevectorSums(CB, D);
       QuickSort(D->codeorder, BookSize(CB), sizeof(int), D->codesum, cmpCodeSum);
       for(i=0; i<BookSize(CB); i++)
       {
          j=D->codeorder[i];
	      D->mynumber[j]=i;
       }
    }
}


static void PrintResult(int Quiet)
{
   if(Quiet>=2)
   {
      printf("\n\nDone preconditions/search: %7.2f",nnccalcstotal != 0LL ?
          (double)npretotal/nnccalcstotal : 0.0);
      printf("\ntauAverage: %7.2f \ndist calcs/search: %7.2f dim/dist calcs: %7.2f dim/seaerch: %7.2f \nksum: %7.0f ndistcalcs: %7.0f searches: %7.0f\n",
      (double)ntautotal/niteration,
      nnccalcstotal != 0LL ? (double)ndistcalcstotal/nnccalcstotal : 0.0,
      ndistcalcstotal != 0LL ? (double)ksumtotal/ndistcalcstotal : 0.0,
      nnccalcstotal != 0LL ? (double)ksumtotal/nnccalcstotal : 0.0,
      (double)ksumtotal,(double)ndistcalcstotal,(double)nnccalcstotal);
        STATCODE(steps=steps1+steps2+steps3+steps4+steps5;
        dists=d1+d2+d3+d4+d5;
        printf("steps1: %7.0f\n",(double)steps1);
        printf("steps2: %7.0f\n",(double)steps2);
        printf("steps3: %7.0f\n",(double)steps3);
        printf("steps4: %7.0f\n",(double)steps4);
        printf("steps5: %7.0f\n",(double)steps5);
        printf("steps: %7.0f\n",(double)steps);
        printf("dist1: %7.0f\n",(double)d1);
        printf("dist2: %7.0f\n",(double)d2);
        printf("dist3: %7.0f\n",(double)d3);
        printf("dist4: %7.0f\n",(double)d4);
        printf("dist5: %7.0f\n",(double)d5);
        printf("dists: %7.0f\n",(double)dists);)
   }
}


static void FreeDataStructures(DATA* D, KNNDATA* KNN, int origCBsize,
KDATA* KD, HEAP* H, int Heap, int Mps)
{
   int i;
   VERTEX* p;
   VERTEX* pe;
   if(Mps)
   {
      deallocate(D->codesum);
      deallocate(D->codeorder);
      deallocate(D->mynumber);
   }
   for( i = 0; i < origCBsize; i++ )
   {
      deallocate(KNN->NN[i]);
   }
   deallocate(KNN->NN);
   for(i = 0; i < origCBsize; i++ )
   {
       p=KNN->head[i];
       while(p!=NULL)
       {
           pe=p;
           p=p->link;
           deallocate(pe);
       }
   }
   deallocate(KNN->head);
   if(Heap)
   {
      for(i=0; i< origCBsize; i++)
      deallocate(KD->DM[i]);
      deallocate(KD->DM);
      Heap_free(&(KD->H));
   }
}


/*-------------------------------------------------------------------*/


void PNNGraph(TRAININGSET*	TS,
			CODEBOOK*	    CB,
            PARTITIONING*	P,
            int           	K,
            int           	FinalSize,
            int             Heap,
            int				GCMethod,
            int           	Quiet,
            long		    watch,
            int				Pds,
            int				Mps,
            int             Mlf,
            int				Lazy,
            int             Cf)
{
  KNNDATA KNN;
  DATA D;
  KDATA KD;
  HEAP H;
  int a,b;
  int origCBsize = BookSize(CB);
  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  CreateDataStructures(CB, &D, &KNN, K, &KD, &H, Heap, Mps);
  InitializeData(CB, &D, Mps);
  switch(GCMethod) /*KGraph creation*/
  {
    case 0:
      /*Brute Force*/
      CalculateAllDistancesGraph(CB, &D, &KNN, K, &KD, &H, Heap, Quiet, watch, Pds);
      break;
    case 1:
      /*Brute Force / fast variant*/
      CalculateAllDistancesGraphFast(CB, &D, &KNN, K, &KD, &H, Heap, Quiet, 
          watch, Pds, Mps, Mlf);
      break;
    case 2:
      /*K Near Neighbors (under construction)*/
      /*CalculateKNNGraph(CB, &D, &KNN, K, Quiet, watch, Pds, Mps);*/
      CalculateAllDistancesGraph(CB, &D, &KNN, K, &KD, &H, Heap, Quiet, watch, Pds);
      break;
    default:
      printf("\nYou may have a wrong GraphCreationMethod option!\n\n");
      break;
  }
  PrintTime(watch);
  PrintProgress(TS, CB, P, Quiet, watch);
  PrintResult(Quiet);
  niteration=0LL;

  while( BookSize(CB) > FinalSize)
  {
     FindMergedPairGraph(CB, &D, &KNN, K, &KD, &H, Heap, &a, &b, Pds, Mps);
	 MergeVectorsGraph(TS, CB, P, &KNN, K, &KD, &H, Heap, a, b, Lazy);
     RecalculateDistancesGraph(CB, &KNN, K, &KD, &H, Heap, a, Quiet, Pds, Lazy);
     PrintProgress(TS, CB, P, Quiet, watch);
     if(ExitRequested()) ExitProcessing(-2);
  }

  PrintResult(Quiet);
  FreeDataStructures(&D, &KNN, origCBsize, &KD, &H, Heap, Mps);
  PrintTime(watch);
}


/* ========== PNNGraph2 ============================================ */


static void PrintList(KNNDATA* KNN, int x)
{
VERTEX *p;
p=KNN->head[x];
while(p!=NULL)
{
    printf("\n%d",p->data);
    p=p->link;
}
}


static void Virhe(char teksti[])
{
    printf("\a%s", teksti);
    exit(1);
}


static void Insert_x_to2list(KNNDATA* KNN, int x, int l)
{
  VERTEX *pi=NULL;
  pi=(VERTEX*)allocate(sizeof(VERTEX));
    if(pi==NULL)Virhe("Muistinvaraus eponnistui");
  pi->data=x;
  pi->link=KNN->head[l];
  KNN->head[l]=pi;
}


static void Remove_x_from2list(KNNDATA* KNN, int x, int l)
{
  VERTEX *pr=NULL, *pe=NULL;
  if(KNN->head[l]==NULL)return;
  pr=KNN->head[l];
  if(pr->data==x)
  {
      STATCODE(steps2++;)
      KNN->head[l]=pr->link;
      deallocate(pr);
      pr=NULL;
  }
  else
  {
    pe=pr;
    pr=pr->link;
    while(pr!=NULL)
    {
      STATCODE(steps2++;)
      if(pr->data==x)
      {
        pe->link=pr->link;
        deallocate(pr);
        pr=NULL;
        break;
        //pr=pe->link;
      }
      else
      {
      pe=pr;
      pr=pr->link;
      }
    } 
  }
}


static int Exists_x_in2list(KNNDATA* KNN, int x, int l)
{
VERTEX *pe=KNN->head[l];
while(pe!=NULL)
{
    if(pe->data==x)
    {
        return 1;
    }
    pe=pe->link;
}
return 0;
}


/*-------------------------------------------------------------------*/


static void AddKNNToLinkedList(CODEBOOK* CB, KNNDATA* KNN, int K)
{
   int i, j, x;
   for(i=0; i<BookSize(CB); i++)
     for(j=0; j<K; j++)
     {
       x=KNN->NN[i][j].Vector;
       if(x!=i)
       {
         Insert_x_to2list(KNN, i, x);
       }
     }
}


/*-------------------------------------------------------------------*/


static void FindMergedPairGraph2(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
 KDATA* KD, HEAP* H, int Heap, int* a, int* b, int Pds, int Mps)
{
  llong MinDist=MAXLLONG, Dist;
  int i;
  int nrec=0;
  DISTELEM* cell;

  if(Heap)
  {
     cell = Heap_getelemdata(&(KD->H), 1, NULL);
     *a = cell->from;
     *b = cell->to;
     STATCODE(steps1++;)
  }

  else /*Linear search and lazy*/
  {
    for( i=0; i<BookSize(CB); i++ )
    {
    STATCODE(steps1++;)
     if( KNN->NN[i][0].VectorDist < MinDist )
        if(KNN->NN[i][0].Recalculate && i!=KNN->NN[i][0].Vector) /*Lazy*/
        {
           Dist=MergeDistortion(CB,i,KNN->NN[i][0].Vector);
           KNN->NN[i][0].VectorDist = Dist;
           KNN->NN[i][0].Recalculate = NO;
           nrec++;
           if(KNN->NN[i][0].VectorDist < MinDist)
           {
              (*a) = i;
              (*b) = KNN->NN[i][0].Vector;
              MinDist = KNN->NN[i][0].VectorDist;
           }
        }
        else
        {
           (*a) = i;
           (*b) = KNN->NN[i][0].Vector;
           MinDist = KNN->NN[i][0].VectorDist;
        }
    }
  }

  /*printf("\nFound %d %d %f BS= %d",*a,*b,(float)KNN->NN[*a][0].VectorDist, BookSize(CB));*/
  STATCODE
      (
      ntau=nrec;
      nnccalcs=nrec;
      )
}

static void FindMergedPairGraph3(CODEBOOK* CB, DATA* D, KNNDATA* KNN, int K,
 KDATA* KD, HEAP* H, int Heap, int* a, int* b, int Pds, int Mps)
{
  int i, j, currcb = 0, *lut;
  llong MinDist=MAXLLONG, Dist;
  int nrec=0;
  DISTELEM* cell;

  lut = (int *) allocate(sizeof(int)*BookSize(CB));

  for ( i = 0; i < BookSize(CB); i++) {
    if (VectorFreq(CB,i) != 0) {
      lut[i] = currcb;
      currcb++;
    } else {
      lut[i] = -1;
    }
  }    


  if(Heap) {
    cell = Heap_getelemdata(&(KD->H), 1, NULL);
    *a = cell->from;
    *b = cell->to;
    printf("\nFound %d %d %f BS= %d",*a,*b,(float)KNN->NN[*a][0].VectorDist, 
	   BookSize(CB));
    return;
  }

  for( i=0; i<BookSize(CB); i++ ) {
    if (KNN->NN[i][0].Vector == i)
      printf("Vertex %d: nan\n", lut[i]);
    else 
      printf("Vertex %d: %d %f %d %f %f\n", lut[i], lut[KNN->NN[i][0].Vector],
	     (float) KNN->NN[i][0].VectorDist, lut[KNN->NN[i][1].Vector], (float) KNN->NN[i][1].VectorDist, (float) MinDist);
    if( KNN->NN[i][0].VectorDist < MinDist ) {
      (*a) = i;
      (*b) = KNN->NN[i][0].Vector;
      MinDist = KNN->NN[i][0].VectorDist;
      //      printf("Vertex %d: %d %f %f\n", i, (*b), 
      //    	     (float) KNN->NN[i][0].VectorDist, (float) MinDist);
    }
  }

  printf("\nFound %d %d %f BS= %d",lut[*a],lut[*b],
	 (float)KNN->NN[*a][0].VectorDist, BookSize(CB));
}


/*-------------------------------------------------------------------*/


static void AddToCC(TRAININGSET*  TS,
                    PARTITIONING* P,
                    int           TSvec,
                    int           Pindex)
{
  int k;
  BOOKNODE* v = &Node(TS, TSvec);

  for( k = 0; k < P->Vsize; k++ )
    {
    CCScalar(P, Pindex, k) += v->freq * (llong)v->vector[k];
    }
  CCFreq(P, Pindex) += v->freq;
  UniqueVectors(P, Pindex)++;
}


/*-------------------------------------------------------------------*/


static void SubtractFromCC(TRAININGSET*  TS,
                           PARTITIONING* P,
                           int           TSvec,
                           int           Pindex)
{
  int k;
  BOOKNODE* v = &Node(TS, TSvec);

  for( k = 0; k < P->Vsize; k++ )
    {
    CCScalar(P, Pindex, k) -= v->freq * (llong)v->vector[k];
    }
  CCFreq(P, Pindex) -= v->freq;
  UniqueVectors(P, Pindex)--;
}


/*-------------------------------------------------------------------*/


static void JoinPartitions2(TRAININGSET* TS, PARTITIONING* P, int To, int From)
{
  int j = FirstVector(P, To);

  if( EndOfPartition(j) )
    {
    FirstVector(P, To) = FirstVector(P, From);
    }
  else
    {
    while( ! IsLast(P, j) )
      {
      j = NextVector(P, j);
      }
    NextVector(P, j) = FirstVector(P, From);
    }
  for( j = FirstVector(P, From); ! EndOfPartition(j); j = NextVector(P, j) )
    {
    Map(P, j) = To;
    SubtractFromCC(TS, P, j, From);
    AddToCC(TS, P, j, To);
    }
  FirstVector(P, From) = ENDPARTITION;

  /* move vectors from the last partition to the empty one */

/*  if ( From != P->PartitionCount-1 )
    {
    while ( ! EndOfPartition(FirstVector(P,P->PartitionCount-1)) )
      {
      ChangePartition(TS, P, From, FirstVector(P, P->PartitionCount-1));
      }
    }

  P->PartitionCount--;*/
}


/*-------------------------------------------------------------------*/


static void MergeVectorsGraph2(TRAININGSET*  TS,
                  CODEBOOK*     CB,
                  int           CBsize,
                  PARTITIONING* P,
                  KNNDATA*      KNN,
                  int           K,
                  KDATA*        KD,
                  HEAP*         H,
                  int           Heap,
                  int           a,
                  int           b,
                  int			  Lazy)
{
  NEARESTTYPE APU[K];
  VERTEX *p=NULL, *pe=NULL;
  llong Dist;
  int i,j,flag;
  int L=K-1;
  int nrec=0;
  /*int last = BookSize(CB)-1;*/
  for( j=0; j<K; j++)
  {
     STATCODE(steps2++;)
     APU[j].VectorDist=KNN->NN[a][j].VectorDist;
     APU[j].Vector=KNN->NN[a][j].Vector;
     APU[j].Recalculate=KNN->NN[a][j].Recalculate;
     KNN->NN[a][j].VectorDist = MAXLLONG/10;
     KNN->NN[a][j].Vector=a;
     KNN->NN[a][j].Recalculate=NO;
  }

  if(Lazy)MarkClustersForRecalculation(CB, KNN, K, a, b);
  JoinPartitions2(TS, P, a, b);
  PartitionCentroid(P, a, &Node(CB, a));

  for( j=0; j<K; j++)/*We check the neighbors of the a*/
  {
     if(APU[j].Vector == b)
     {
         Remove_x_from2list(KNN, a, b);
         continue;
     }
     if(APU[j].Vector == a)continue;
     flag=0;
     for(i=0; i<K; i++)
     {
         STATCODE(steps2++;)
         if(APU[j].Vector==KNN->NN[a][i].Vector)
         {
             flag=1;
             continue;
         }
     }
     if(flag)
     {
         //printf("\n?! Two pointers for same vector a");
         continue;
     }
     STATCODE(d2++;)
     Dist = MergeDistortion(CB, a, APU[j].Vector);
     nrec++;
     if(Dist<KNN->NN[a][L].VectorDist)
     {
   	    SaveNeighbor(KNN, K, Dist, a, APU[j].Vector);
     }
     else
     {
         /*Not possible!!*/
         printf("\nNot OK!");
     }
  }/*End for*/

  for( j=0; j<K; j++)/*We check the neighbors of the b*/
  {
     Remove_x_from2list(KNN, b, KNN->NN[b][j].Vector);
     if(KNN->NN[b][j].Vector == a)
     {
         KNN->NN[b][j].Vector = b;
         KNN->NN[b][j].VectorDist= MAXLLONG/10;
         KNN->NN[b][j].Recalculate=NO;
         continue;
     }
     if(KNN->NN[b][j].Vector == b)continue;
     flag=0;
     for(i=0; i<K; i++)
     {
         STATCODE(steps2++;)
         if(KNN->NN[b][j].Vector==KNN->NN[a][i].Vector)
         {
           flag=1;
           KNN->NN[b][j].Vector = b;
           KNN->NN[b][j].VectorDist= MAXLLONG/10;
           KNN->NN[b][j].Recalculate=NO;
           continue;
         }
     }
     if(flag)
     {
         continue;
     }
     STATCODE(d2++;)
     Dist = MergeDistortion(CB, a, KNN->NN[b][j].Vector);
     nrec++;
     if(Dist<KNN->NN[a][L].VectorDist)
     {
        if(KNN->NN[a][L].Vector != a)
        {
            /*We remove a from the second list of the KNN->NN[a][L].Vector*/
            Remove_x_from2list(KNN, a, KNN->NN[a][L].Vector);
        }
        SaveNeighbor(KNN, K, Dist, a, KNN->NN[b][j].Vector);
        /*We insert a to the second list of the KNN->NN[b][j].Vector*/
        if(!Exists_x_in2list(KNN, a, KNN->NN[b][j].Vector))
        Insert_x_to2list(KNN, a, KNN->NN[b][j].Vector);
        KNN->NN[b][j].Vector = b;
        KNN->NN[b][j].VectorDist= MAXLLONG/10;
        KNN->NN[b][j].Recalculate=NO;
     }
     else
     {
         KNN->NN[b][j].Vector = b;
         KNN->NN[b][j].VectorDist= MAXLLONG/10;
         KNN->NN[b][j].Recalculate=NO;
     }
  }/*End for*/

  if(Heap)
  {
     Heap_remove(&(KD->H), DM_heapindex(KD, b), NULL);
     STATCODE(steps4=steps4+round(log10(CBsize)/log10(2.));)
     DM_distset(KD, a, KNN->NN[a][0].VectorDist);
     DM_toset(KD, a, KNN->NN[a][0].Vector);
     Heap_update(&(KD->H),DM_heapindex(KD, a),NULL);
     STATCODE(steps2=steps2+round(log10(CBsize)/log10(2.));)
  }

  p=KNN->head[b]; /*We check bottom*/
  KNN->head[b]=NULL;
  while(p!=NULL)
  {
    i=p->data;
    //if(i==a)printf("\ni=a???");
  	 if(!Exists_x_in2list(KNN, i, a))
  	 {
        Insert_x_to2list(KNN, i, a);
  	 }
  	 for(j=0; j<K; j++)
  	 {
       STATCODE(steps3++;)
   	  if(KNN->NN[i][j].Vector==b)
      {
      	KNN->NN[i][j].Vector=a;/*Voi jo olla KNN->NN[i][j].Vector==a?!*/
        if(Heap && j==0) DM_toset(KD, i, KNN->NN[i][0].Vector);
      }
  	 }

  	 pe=p;
  	 p=p->link;
  	 deallocate(pe);
    pe=NULL;
  }

  VectorFreq(CB,b)=0;
  STATCODE(steps4++;)
  
  STATCODE
      (
      ntau=nrec;
      nnccalcs=nrec;
      )
}


/*-------------------------------------------------------------------*/


static void RecalculateDistancesGraph2(CODEBOOK* CB, int CBsize, KNNDATA* KNN, int K,
               KDATA* KD, HEAP* H, int Heap, int aex, int Pds, int Lazy)
{
  VERTEX *p=NULL;
  int  i,j;
  STATCODE(int  nrec=0;)
  llong Dist;

  if(Lazy)return;

  j=0;
  p=KNN->head[aex];
  while(p!=NULL)
  {
      STATCODE(steps5++;)
  	   i=p->data;/*i things that aex is among k near neignbor*/
      if(KNN->NN[i][j].Vector==aex && (KNN->NN[i][j].Vector!=i))
      {
         STATCODE(d5++;)
         Dist=MergeDistortion(CB, i, aex);
         KNN->NN[i][j].VectorDist=Dist;
         STATCODE(nrec++;)
         if(Heap)
         {
         DM_distset(KD, i, KNN->NN[i][j].VectorDist);
         Heap_update(&(KD->H), DM_heapindex(KD, i), NULL);
         STATCODE(steps5=steps5+round(log10(CBsize)/log10(2.));)
         }
      }
      KNN->NN[i][j].Recalculate = NO;
      p=p->link;
  }

  STATCODE
      (
      nnccalcs+=nrec;
      ntau+=nrec;
      )
}


/*-------------------------------------------------------------------*/

/* forms the graph to g  */
Graph * PNNGraph2(TRAININGSET*	    TS,
	       CODEBOOK*	    CB,
               PARTITIONING*	P,
	       int           	K,
               int           	FinalSize,
               int              Heap,
               int				GCMethod,
               int           	Quiet,
               long			    watch,
               int				Pds,
               int				Mps,
               int              Mlf,
               int				Lazy,
               int              Cf)
{
  KNNDATA KNN;
  DATA D;
  KDATA KD;
  HEAP H;
  Graph* g;
  int a,b, i = 1;
  int origCBsize = BookSize(CB);
  int CBsize = BookSize(CB);
  char *bufferi;

  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  CreateDataStructures(CB, &D, &KNN, K, &KD, &H, Heap, Mps);
  InitializeData(CB, &D, Mps);
  switch(GCMethod) /*KNNGraph creation*/
  {
    case 0:
      /*Brute Force*/
      CalculateAllDistancesGraph(CB, &D, &KNN, K, &KD, &H, Heap, Quiet, watch, Pds);
      break;
    case 1:
      /*Brute Force / fast variant*/
      CalculateAllDistancesGraphFast(CB, &D, &KNN, K, &KD, &H, Heap, Quiet,
          watch, Pds, Mps, Mlf);
      break;
    case 2:
      /*Divide and Conquer*/
      /*CalculateAllDistancesGraphDaC(CB, &KNN, K, &KD, &H, Heap, Quiet,
          watch, Pds, Cf); keskenerinen */
      CalculateAllDistancesGraphFast(CB, &D, &KNN, K, &KD, &H, Heap, Quiet,
          watch, Pds, Mps, Mlf);
      break;
    default:
      printf("\nYou may have a wrong GraphCreationMethod option!\n\n");
      break;
  }
  // PnnGraphWriteGraph("start.dgr", CB, &KNN, K, BookSize(CB));
  printf("Foobar\n");
  AddKNNToLinkedList(CB, &KNN, K);
  PrintTime(watch);
  PrintProgress(TS, CB, P, Quiet, watch);
  PrintResult(Quiet);
  niteration=0LL;

  bufferi = (char *) allocate(sizeof(char)*256);

  while( CBsize > FinalSize)
  {
     if(Quiet>=2)printf("\n BS= %i",CBsize);
     if (CBsize == FinalSize+1)
       FindMergedPairGraph3(CB, &D, &KNN, K, &KD, &H, Heap, &a, &b, Pds, Mps);
     else
       FindMergedPairGraph2(CB, &D, &KNN, K, &KD, &H, Heap, &a, &b, Pds, Mps);
     MergeVectorsGraph2(TS, CB, CBsize, P, &KNN, K, &KD, &H, Heap, a, b, Lazy);
     RecalculateDistancesGraph2(CB, CBsize, &KNN, K, &KD, &H, Heap, a, Pds, Lazy);
     CBsize--;
     PrintProgress(TS, CB, P, Quiet, watch);
     if(ExitRequested()) ExitProcessing(-2);
     
     //     sprintf(bufferi, "iteration-%d.dgr", i++);
     //    PnnGraphWriteGraph(bufferi, CB, &KNN, K, CBsize);
  }

  PrintResult(Quiet);
  /* the next changed by mm */
  g = PnnGraphFormGraph(CB, &KNN, K, FinalSize);
  //  PnnGraphWriteGraph("final.dgr", CB, &KNN, K, FinalSize);
  FreeDataStructures(&D, &KNN, origCBsize, &KD, &H, Heap, Mps);
  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  PrintTime(watch);
  return g;
}
